from enum import Enum
from typing import TypeVar, Any, List, Union, Dict

######### Common ###########
MessageDict = Dict[str, Any]
MessageDictOrStr = Union[str, MessageDict]
SupportedMessages = Union[MessageDictOrStr, List[MessageDictOrStr]]
Output = TypeVar('Output')

class Role(str, Enum):
    system = "system"
    user = "user"
    assistant = "assistant"
    tool = "tool"
    # make it similar to Literal["system", "user", "assistant", tool] while retaining enum convenience

    def __new__(cls, value, *args, **kwargs):
        obj = str.__new__(cls, value)
        obj._value_ = value
        return obj

    def __str__(self):
        return str(self.value)
    





from typing import Callable, Dict, List

class SingletonMeta(type):
    """
    A metaclass that creates a Singleton base type when called.
    """
    _instances: Dict[type, object] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

SubscriberCallback = Callable[[...],None]

class JustEventBus(metaclass=SingletonMeta):
    """
    A simple singleton event bus to publish function call results for functions that support it
    Event name can be anything, but suggested use is function names.
    """

    _subscribers: Dict[str, List[SubscriberCallback]]

    def __init__(self):
        # Dictionary of subscription_pattern -> list_of_callbacks (in subscription order)
        self._subscribers = {}

    def subscribe(self, event_prefix: str, callback: SubscriberCallback):
        """Subscribe a callback to a specific event."""
        """
        Subscribe to an event name or prefix:
          e.g. 'mytool.call' or 'mytool.*'
        """
        if event_prefix not in self._subscribers:
            self._subscribers[event_prefix] = []
        self._subscribers[event_prefix].append(callback)

    def publish(self, event_name: str, *args, **kwargs):
        """
        Publish an event to:
          1) Any exact-match subscriber (same string)
          2) Any prefix subscriber (ends with '.*' and event_name.startswith(prefix))
        """
        # (1) Get exact-match subscribers
        exact_callbacks = self._subscribers.get(event_name, [])

        # (2) Find prefix-matching subscribers
        prefix_callbacks : List[SubscriberCallback]= []
        for pattern, callbacks in self._subscribers.items():
            if pattern.endswith('.*'):
                prefix = pattern[:-2]
                if event_name.startswith(prefix):
                    prefix_callbacks.extend(callbacks)

        ## Invoke the callbacks with deduplication ( preserve order of subscription)
        seen = set()
        for cb in exact_callbacks + prefix_callbacks:
            if cb not in seen:
                cb(event_name, *args, **kwargs)
                seen.add(cb)









from just_agents.interfaces.streaming_protocol import IAbstractStreamingProtocol
import json
import time

class OpenaiStreamingProtocol(IAbstractStreamingProtocol):
    def get_chunk(self, index: int, delta: str, options: dict):
        chunk = {
            "id": index,
            "object": "chat.completion.chunk",
            "created": time.time(),
            "model": options["model"],
            "choices": [{"delta": {"content": delta}}],
        }
        return f"data: {json.dumps(chunk)}\n\n"


    def done(self):
        # https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#event_stream_format
        return "data: [DONE]\n\n"